/* tapconv - Spectrum emulator tape-file converter
 *
 * Public Domain software
 *
 * Author: Thomas Ahn Kjaer
 *
 * Convert one tape-file to another (or +3DOS files)
 *
 * usage:
 *        tapconv <infile> <outfile> <outformat>
 *
 *                         outformat = tap, z80v2.01
 *                                     xzx, xz80 or xzx
 *                                     p3d, +3DOS (Spectrum +3)
 *                                     raw, data only (eg. z80v2.01 screen)
 *
 * example:
 *         convert a z80v2.01 .TAP file to +3DOS file.
 *         tapconv z80v201file.tap outfile p3d
 *
 * As there's no way to detect the z80 .TAP format, it relies on the extension
 * used by the z80 v2.x/3.x emulator .TAP (or .tap).
 * If neither of tap, xzx, p3d is detected, the raw format is assumed.
 */

#define DEBUG

/* defines */
#define TAP 200
#define XZX 201
#define P3D 202
#define RAW 203

static char XZXID[4] = "ZXF1";
static char PLUS3ID[8] = "PLUS3DOS";
static char PLUS3PREF[10] = {'P','L','U','S','3','D','O','S',0x1a,0x01};

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(__EMX__) || defined(MSDOS)
#include <sys/types.h>
#endif
#include <sys/stat.h>
#include <ctype.h>

/* MUST be 8 bits! */
typedef unsigned char uns8;
typedef unsigned short uns16;

/* the header's */
struct raw_header {
  uns8 type;
  uns8 name[10];
  uns8 lenlow;
  uns8 lenhigh;
  uns8 parm1low;
  uns8 parm1high;
  uns8 parm2low;
  uns8 parm2high;
} raw_hd;

struct tap_header2 {
  uns8 blklenlow;
  uns8 blklenhigh;
  uns8 flagA;
} tap2_hd;

struct tap_header {		/* header layout */
  struct raw_header h;		/* same as for raw */
  uns8 chksum;
} tap_hd;

struct xzx_header {
  uns8 xzxid[4];
  struct raw_header h;		/* same as for raw */
} xzx_hd;

struct p3d_header {
  uns8 prefix[10];
  uns8 version;			/* 10    */
  uns8 totlenlow;		/* 11    */
  uns8 totlenhigh;		/* 12    */
  uns8 totlenhl;		/* 13    */
  uns8 totlenhh;		/* 14    */
  uns8 type;			/* 15    */
  uns8 lenlow;			/* 16    */
  uns8 lenhigh;			/* 17    */
  uns8 parm1low;
  uns8 parm1high;
  uns8 parm2low;
  uns8 parm2high;		/* 21    */
  uns8 dump4[105];		/* 126   */
  uns8 chksum;			/* 127   */
} p3d_hd;

FILE *fpIn;
FILE *fpOut;
uns8 data_block[49152];   /* I wanted 65536, but Russell only 49152 :-) */
size_t file_size;         /* for raw files */
uns16 tap_blocks;         /* number of blocks in .TAP */
uns16 tap_sel;            /* selected block number in .TAP */
long tap_pos[256];        /* position of blocks, no more than 256 blocks! */

/* prototypes */
void usage(void);
void print_type(int);
void print_header(void);
int determine_intype(char *, char *);
int determine_output(char *);
void read_header(int);
void read_tap_header(void);
void read_xzx_header(void);
void read_p3d_header(void);
void read_raw_header(void);
void prepare_header(int, int);
void write_all(int, long, char *);
uns16 examine_tap(void);


int main(int argc, char *argv[])
{
  int intype;
  int outtype;
  
  struct stat status;
  long block_len;		/* length of data block */
  uns16 hd_len;			/* length as seen by the header */
  char checkid[8];		/* +3DOS and XZX files starts with an string ID */
  char *szFrom, *szTo;

  fprintf(stderr, "%d\n", sizeof(struct p3d_header));
  if (argc != 4)
    usage();

  /* can't handle wildcard, for MS-DOG & OS/2 and probaly other?? */
  if ((strchr(argv[1],'*')!=NULL) || (strchr(argv[1],'?')!=NULL)) {
    fprintf(stderr, "This program can't handle wildcards, sorry!\n");
    return 1;
  }
  
  /* does the input file exists? */
  if(stat(argv[1], &status)<0) {
    perror(argv[1]);
    return 1;
  }

  file_size = status.st_size;

  if ((szFrom = malloc(strlen(argv[1])))==NULL){
    perror(argv[1]);
    return 1;
  };
  strcpy(szFrom, argv[1]);

  if ((szTo = malloc(strlen(argv[2])))==NULL){
    perror(argv[2]);
    return 1;
  };
  strcpy(szTo, argv[2]);

  /* try to determine the input format */  
  
  if ((fpIn = fopen(szFrom, "rb"))==NULL){
    perror(szFrom);
    return 1;
  };

  fscanf(fpIn, "%8s", checkid);
  intype = determine_intype(checkid, szFrom);
  outtype = determine_output(argv[3]);

#ifdef DEBUG
  fprintf(stderr, "Converting %s from ", szFrom);
  print_type(intype);
  fprintf(stderr, " to ");
  print_type(outtype);
  fprintf(stderr, "\n");
#endif
  if (intype==outtype) {
    fprintf(stderr, "intype and outtype are the same, no conversion done!\n");
    fclose(fpIn);
    exit(1);
  }

  if (intype==TAP){
    tap_blocks = examine_tap();
    while(1){
      fprintf(stderr, "select block to save: ");
      scanf("%u", (unsigned int *)&tap_sel);
      if (tap_sel <= tap_blocks) break;
      fprintf(stderr, "\007Error, wrong block number\n");
    }
  }

  read_header(intype);

  prepare_header(outtype, intype);

  /* read data block from infile */
  hd_len = raw_hd.lenhigh*256 + raw_hd.lenlow;
  block_len = fread(&data_block[0], (size_t)sizeof(uns8), (size_t)hd_len, fpIn);
  if (hd_len != block_len) {
    fprintf(stderr, "\nerror reading data block!!\n");
    fclose(fpIn);
    exit(1);
  }
  
  write_all(outtype, hd_len, szTo);

  fclose(fpIn);
  return 0;
}


/* determine the input format */
int determine_intype(char *id, char *szName)
{
  char *p;
  char ext[5];
  int i;

  if (strncmp(XZXID, id, 4)==0)   return XZX;
  if (strncmp(PLUS3ID, id, 8)==0) return P3D;
  
  /* o.k type can't be detected on the ID, try .TAP */
  p = strrchr(szName, '.');
  if (p==NULL) 
    return RAW;			/* no dot in filename, assume RAW */
  else {
    for (i = 0; i < 4; i++){
      if (*p != '\0') {
	ext[i] = toupper(*p);
	p++;
      } 
      else {
	ext[i] = '\0';
      }
    }
    ext[4] = '\0';
    
    if (strcmp(ext, ".TAP")==0)
      return TAP;
  }
  
  return RAW;			/* if the extension was not .TAP */
}


/* determine output format */
int determine_output(char *type)
{
  char ext[5];
  char *p;
  int i;

  p = type;
  for (i = 0; i < 4; i++) {
    if (*p != '\0') {
      ext[i] = toupper(*p);
      p++;
    }
    else
      ext[i] = '\0';
  }

  if (strcmp(ext, "TAP")==0) return TAP;
  if (strcmp(ext, "XZX")==0) return XZX;
  if (strcmp(ext, "P3D")==0) return P3D;
  if (strcmp(ext, "RAW")==0) return RAW;
  
  fprintf(stderr, "\n\ninvalid output type\n");
  fclose(fpIn);
  usage();
  return -1;
}


/* read_header - reads header from infile and prepares the structure header */
void read_header(int type)
{
  rewind(fpIn);
  switch(type){
  case TAP:
    read_tap_header();
    break;
  case XZX:
    read_xzx_header();
    break;
  case P3D:
    read_p3d_header();
    break;
  case RAW:
    read_raw_header();
    break;
  default:
    break;
  }

#ifdef DEBUG
  print_header();
#endif

}

/* examines the .TAP file and count the number of blocks */
uns16 examine_tap()
{
  int no;
  uns16 len;

  rewind(fpIn);
  no = 0;
  while (1) {
    tap_pos[no] = ftell(fpIn);
    if ((fread(&tap2_hd, sizeof(struct tap_header2), 1, fpIn))!=1){
      rewind(fpIn);
      return (no -1 );
    }
    len = (tap2_hd.blklenhigh * 256 + tap2_hd.blklenlow) - 2;
    fprintf(stderr, "BLOCK %d: ", no);
    switch(tap2_hd.flagA){
    case 0:
      fprintf(stderr, "header\n");
      if ((fread(&tap_hd, sizeof(struct tap_header), 1, fpIn))!=1){
	fprintf(stderr, "error reading header\n");
	fclose(fpIn);
	exit(1);
      }
      break;
    case 255:
      fprintf(stderr, "  data, len=%d\n", len);
      if ((fread(&data_block[0], sizeof(uns8), len, fpIn))!=len){
	fprintf(stderr, "error reading data block\n");
	fclose(fpIn);
	exit(1);
      }
      
      if ((fread(&data_block[0], sizeof(uns8), 1, fpIn))!=1){
	fprintf(stderr, "error reading checksum\n");
	fclose(fpIn);
	exit(1);
      }
      break;
    default:
      fprintf(stderr, "unknown flag type!\n");
      fclose(fpIn);
      exit(1);
      break;
    }
    no++;
  }
  
  return no;
}


void read_tap_header()
{
  fseek(fpIn, tap_pos[tap_sel], SEEK_SET);
  if ((fread(&tap2_hd, sizeof(struct tap_header2), 1, fpIn))!=1){
    fprintf(stderr, "error reading header\n");
    fclose(fpIn);
    exit(1);
  }
  if ((fread(&tap_hd, sizeof(struct tap_header), 1, fpIn))!=1){
    fprintf(stderr, "error reading header\n");
    fclose(fpIn);
    exit(1);
  }
  if ((fread(&tap2_hd, sizeof(struct tap_header2), 1, fpIn))!=1){
    fprintf(stderr, "error reading header\n");
    fclose(fpIn);
    exit(1);
  }
  raw_hd = tap_hd.h;
}

void read_xzx_header()
{
  if ((fread(&xzx_hd, sizeof(struct xzx_header), 1, fpIn))!=1){
    fprintf(stderr, "error reading header\n");
    fclose(fpIn);
    exit(1);
  }
  raw_hd = xzx_hd.h;
}

void read_p3d_header()
{
  if ((fread(&p3d_hd, sizeof(struct p3d_header), 1, fpIn))!=1){
    fprintf(stderr, "error reading header\n");
    fclose(fpIn);
    exit(1);
  }
  raw_hd.type = p3d_hd.type;
  raw_hd.lenlow = p3d_hd.lenlow;
  raw_hd.lenhigh = p3d_hd.lenhigh;
  raw_hd.parm1low = p3d_hd.parm1low;
  raw_hd.parm1high = p3d_hd.parm1high;
  raw_hd.parm2low = p3d_hd.parm2low;
  raw_hd.parm2high = p3d_hd.parm2high;
}

void read_raw_header()
{
  long start;

  /* the user has to specify the layout of the header */
  fprintf(stderr, "\nraw file has length %d\n", (int)file_size);
  fprintf(stderr, "Enter start address: ");
  scanf("%d", (int *)&start);
  raw_hd.type = 3;		/* code */
  raw_hd.lenlow = file_size % 256;
  raw_hd.lenhigh = file_size / 256;
  raw_hd.parm1low = start % 256;
  raw_hd.parm1high = start / 256;
  raw_hd.parm2low = 0;
  raw_hd.parm2high = 0;
}


void prepare_header(int outtype, int readtype)
{
  char filename[11];
  uns8 *p;
  long i;
  unsigned int chksum;
  
  if ((readtype==P3D)||(readtype==RAW)) {
    /* the user has to specify the filename */
    strncpy(raw_hd.name, "          ", 10);
    fprintf(stderr, "type filename as seen from save\n");
    scanf("%10s", (char *)&filename);

    i = 0;
    p = &filename[0];
    while (*p != '\0') {
      raw_hd.name[i++] = *p++;
    }

  };

  switch(outtype) {
  case TAP:
    /* header */
    tap2_hd.blklenlow = 19;
    tap2_hd.blklenhigh = 0;
    tap2_hd.flagA = 0;
    tap_hd.h = raw_hd;
    p = (uns8 *)&tap_hd.h;
    tap_hd.chksum = tap2_hd.flagA;
    for (i = 0; i < 17; i++)
      tap_hd.chksum ^= *p++;
    break;
  case XZX:
    strncpy(xzx_hd.xzxid, XZXID, 4);
    xzx_hd.h = raw_hd;
    break;
  case P3D:
    strncpy((char *)&p3d_hd.prefix, (char *)&PLUS3PREF, 10);
    i = raw_hd.lenhigh * 256 + raw_hd.lenlow + 128;
    p3d_hd.totlenlow = i % 256;
    p3d_hd.totlenhigh = i / 256;
    p3d_hd.type = raw_hd.type;
    p3d_hd.lenlow = raw_hd.lenlow;
    p3d_hd.lenhigh = raw_hd.lenhigh;
    p3d_hd.parm1low = raw_hd.parm1low;
    p3d_hd.parm1high = raw_hd.parm1high;
    p3d_hd.parm2low = raw_hd.parm2low;
    p3d_hd.parm2high = raw_hd.parm2high;
    chksum = 0;
    p = (uns8 *)&p3d_hd;
    for (i = 0; i < 127; i++)
      chksum += *p++;
    p3d_hd.chksum = chksum % 256;
    break;
  case RAW:
    break;
  default:
    exit(1);
  }

}


void write_all(int outtype, long hd_len, char *szName)
{
  uns16 i;
  uns8 chksum;
  long tot_len;
  
  if ((fpOut = fopen(szName, "wb"))==NULL){
    perror(szName);
    fclose(fpIn);
    fclose(fpOut);
    exit(1);
  };

  switch(outtype) {
  case TAP:
    /* write header */
    fwrite(&tap2_hd, sizeof(struct tap_header2), 1, fpOut);
    fwrite(&tap_hd, sizeof(struct tap_header), 1, fpOut);

    /* prepare for data-block */
    tot_len = hd_len + 2;
    tap2_hd.blklenlow = (tot_len) % 256;
    tap2_hd.blklenhigh = (tot_len) / 256;
    tap2_hd.flagA = 255;
    fwrite(&tap2_hd, sizeof(struct tap_header2), 1, fpOut);
    fwrite(&data_block[0], (size_t)sizeof(uns8), (size_t)hd_len, fpOut);
    
    /* calculate checksum */
    chksum = tap2_hd.flagA;
    for (i = 0; i < hd_len; i++)
      chksum ^= data_block[i];
    fputc(chksum, fpOut);
    break;
  case XZX:
    fwrite(&xzx_hd, sizeof(struct xzx_header), 1, fpOut);
    fwrite(&data_block[0], (size_t)sizeof(uns8), (size_t)hd_len, fpOut); 
    break;
  case P3D:
    fwrite(&p3d_hd, sizeof(struct p3d_header), 1, fpOut);
    fwrite(&data_block[0], (size_t)sizeof(uns8), (size_t)hd_len, fpOut); 
    break;
  case RAW:
    fwrite(&data_block[0], (size_t)sizeof(uns8), (size_t)hd_len, fpOut);
    break;
  }

  fclose(fpOut);
}


/* usage() - print usage and exit */
void usage()
{
  fprintf(stderr, "TAPCONV version 0.1 \n\n");
  fprintf(stderr, "usage: tapconv <infile> <outfile> <outformat>\n\n");
  fprintf(stderr, "infile must be a valid TAP, XZX, P3d or RAW file.\n");
  fprintf(stderr, "outformat must be either TAP, XZX, P3D or RAW.\n");
  fprintf(stderr, "the output is written to <outfile>\n\n");
  fprintf(stderr, "example:\n");
  fprintf(stderr, "\ttapconv a_z80_tapfile.tap newname XZX\n\n");
  fprintf(stderr, "Public Domain, Thomas Ahn Kjaer 1995\n\n");

  exit(1);
}

void print_type(int type)
{
  switch(type){
  case TAP:
    fprintf(stderr, "TAP");
    break;
  case XZX:
    fprintf(stderr, "XZX");
    break;
  case P3D:
    fprintf(stderr, "PLUS3DOS");
    break;
  case RAW:
    fprintf(stderr, "RAW");
    break;
  default:
    fprintf(stderr, "unknown type!! %d", type);
    break;
  }
}

void print_header()
{
  switch(raw_hd.type){
  case 0:
    fprintf(stderr, "Program: ");
    fprintf(stderr, "%.10s\n", raw_hd.name);
    fprintf(stderr, "Len: %u\n", raw_hd.lenhigh*256+raw_hd.lenlow);
    if (raw_hd.parm1high*256+raw_hd.parm1low < 32768)
      fprintf(stderr, "Line: %u\n", raw_hd.parm1high*256+raw_hd.parm1low);
    fprintf(stderr, "BAS-len: %u\n", raw_hd.parm2high*256+raw_hd.parm2low);
    break;
  case 1:
    fprintf(stderr, "NumArray: ");
    fprintf(stderr, "%.10s\n", raw_hd.name);
    fprintf(stderr, "Len: %u\n", raw_hd.lenhigh*256+raw_hd.lenlow);
    break;
  case 2:
    fprintf(stderr, "ChrArray: ");
    fprintf(stderr, "%.10s\n", raw_hd.name);
    fprintf(stderr, "Len: %u\n", raw_hd.lenhigh*256+raw_hd.lenlow);
    break;
  case 3:
    fprintf(stderr, "Code: ");
    fprintf(stderr, "%.10s\n", raw_hd.name);
    fprintf(stderr, "Len: %u\n", raw_hd.lenhigh*256+raw_hd.lenlow);
    fprintf(stderr, "Start: %u\n", raw_hd.parm1high*256+raw_hd.parm1low);
    break;
  default:
    fprintf(stderr, "Unknown: ");
    fprintf(stderr, "%.10s\n", raw_hd.name);
    fprintf(stderr, "Len: %u\n", raw_hd.lenhigh*256+raw_hd.lenlow);
    break;
  }
}
